"""
Rough draft of a script to get the total and free space of all physical drives on a system.
"""

import logging
from typing import Any

import psutil

LOGGER = logging.getLogger(__name__)


def convert_bytes_to_gb(bytes_value: int) -> str:
    """
    Convert bytes to gigabytes.

    Args:
        bytes_value (int): The value in bytes.

    Returns:
        str: The value in gigabytes, formatted as a string with 2 decimal places.
    """
    gb_value = bytes_value / (1024**3)
    return f"{gb_value:.2f} GB"


ignore_fs_types = {
    "proc",
    "sysfs",
    "tmpfs",
    "devtmpfs",
    "devfs",
    "squashfs",
    "cgroup",
    "cgroup2",
    "overlay",
    "none",
    "autofs",
    "rpc_pipefs",
    "hugetlbfs",
    "tracefs",
    "mqueue",
    "securityfs",
    "debugfs",
    "binfmt_misc",
    "configfs",
    "efivarfs",
    "fuse.gvfsd-fuse",
    "fuse.gvfs-fuse-daemon",
    "fusectl",
    "pstore",
    "sys_kernel_debug",
    "sys_kernel_config",
}


def get_free_percent_for_all_drives() -> float:
    """
    Get the percent of free space for all physical drives on the system.

    Returns:
        float: The percent of free space.
    """
    free = 0.0
    total = 0.0
    for data in get_drive_info(use_numbers=True):
        # get percent free
        free += data["Free Space"]
        total += data["Total Space"]
    if total > 0.0:
        return (free / total) * 100
    return 0.0


def get_drive_info(use_numbers: bool = False) -> list[dict[str, Any]]:
    """
    Get a list of all physical mounted drives with their total and free space, excluding virtual or system mounts.

    Returns:
        list[dict[str, Any]]: A list of dictionaries each containing the mount point, total space, and free space.
    """
    drives = []
    for partition in psutil.disk_partitions():
        if partition.fstype not in ignore_fs_types:
            try:
                usage = psutil.disk_usage(partition.mountpoint)
                drive = {
                    "Mount Point": partition.mountpoint,
                    "Total Space": convert_bytes_to_gb(usage.total) if not use_numbers else usage.total,
                    "Free Space": convert_bytes_to_gb(usage.free) if not use_numbers else usage.free,
                }
                drives.append(drive)
            except Exception:
                # Too noisy.
                pass
    return drives


if __name__ == "__main__":
    print(get_free_percent_for_all_drives())



"""
CLI interface. At the moment, the CLI is for demonstration purposes only. Use the decorator or context manager in your
own code.
"""

import argparse
import logging
import logging.config
import sys
import time
from typing import Optional, Sequence, cast

from sparkle_log.__about__ import __version__
from sparkle_log.as_context_manager import MetricsLoggingContext
from sparkle_log.as_decorator import monitor_metrics_on_call
from sparkle_log.ui import GraphStyle


@monitor_metrics_on_call(("cpu",), 1)
def log_memory_and_cpu():
    """Example with decorator"""
    while True:
        time.sleep(4)
        print("Executing demo...")


def log_memory_and_cpu_cli(
    metrics=("cpu", "memory", "drive"), interval: int = 1, duration: int = 10, style: GraphStyle = "bar"
):
    """
    Log memory and CPU metrics using the Sparkle Log system.
    """
    configure_logging()
    with MetricsLoggingContext(metrics=metrics, interval=interval, style=style):
        # Execute some operations
        print("Demo of Sparkle Monitoring system metrics during operations...")
        time.sleep(int(duration / 3))
        print("Maybe CPU intensive work done here...")
        time.sleep(int(duration / 3))
        print("Maybe Memory intensive work done here...")
        time.sleep(int(duration / 3))


def configure_logging() -> None:
    """Turn on color logging"""
    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "colored": {
                    "()": "colorlog.ColoredFormatter",
                    "format": "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
                }
            },
            "handlers": {
                "default": {
                    "level": "DEBUG",
                    "formatter": "colored",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",  # Default is stderr
                },
            },
            "loggers": {
                "sparkle_log": {
                    "handlers": ["default"],
                    "level": "DEBUG",
                    "propagate": False,
                }
            },
        }
    )


def main(argv: Optional[Sequence[str]] = None) -> int:
    """Arg parse code"""
    parser = argparse.ArgumentParser(description="Monitor system metrics using Sparkle Log.")
    parser.add_argument(
        "--metrics",
        type=str,
        default="cpu,memory,drive",
        help="Comma-separated list of metrics to monitor (e.g., 'cpu,memory,drive')",
    )
    parser.add_argument("--interval", type=int, default=1, help="Interval in seconds between metric logs")
    parser.add_argument("--duration", type=int, default=10, help="Duration in sections to gather metrics")
    # An add_argument call with a choice of bar, faces
    parser.add_argument(
        "--style",
        choices=["bar", "faces", "jagged", "linear", "vertical", "ascii_art", "pie_chart"],
        default="bar",
        help="Graph Style",
    )

    parser.add_argument("--version", action="version", version=f"Sparkle Log {__version__}")

    args = parser.parse_args(argv)

    # Convert comma-separated string to tuple of metrics
    metrics_tuple = tuple(args.metrics.split(","))

    # Call the function with parsed arguments
    log_memory_and_cpu_cli(
        metrics=metrics_tuple, interval=args.interval, duration=args.duration, style=cast(GraphStyle, str(args.style))
    )
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



"""
This module contains the decorator to monitor the metrics of the system while the function is being executed.
"""

import logging
from functools import wraps
from threading import Event, Thread

from sparkle_log.graphs import GLOBAL_LOGGER
from sparkle_log.scheduler import run_scheduler
from sparkle_log.ui import GraphStyle

INITIALIZED = False


def monitor_metrics_on_call(metrics=("cpu", "memory"), interval=10, style: GraphStyle = "bar"):
    """
    Decorator to monitor the system metrics while the function is being executed.
    """

    def decorator(func):
        """Wrapper function"""

        @wraps(func)
        def wrapper(*args, **kwargs):
            """Wrapper function"""
            if not GLOBAL_LOGGER.isEnabledFor(logging.INFO):
                return func(*args, **kwargs)
            stop_event = Event()
            scheduler_thread = Thread(target=run_scheduler, args=(stop_event, metrics, interval, style))
            scheduler_thread.start()

            try:
                return func(*args, **kwargs)
            finally:
                # Stop the scheduler and wait for thread to finish
                stop_event.set()

                scheduler_thread.join()

        return wrapper

    return decorator



"""Metadata for sparkle-log."""

__all__ = [
    "__title__",
    "__version__",
    "__description__",
    "__author__",
    "__author_email__",
    "__keywords__",
    "__status__",
    "__license__",
    "__readme__",
    "__repository__",
    "__homepage__",
    "__documentation__",
]

__title__ = "sparkle-log"
__version__ = "0.3.0"
__description__ = "Write sparkline graphs of CPU and memory usage to your logs."
__author__ = "Matthew Martin"
__author_email__ = "matthewdeanmartin@gmail.com"
__keywords__ = ["monitoring", "logging"]
__status__ = "4 - Beta"
__license__ = "MIT"
__readme__ = "README.md"
__repository__ = "https://github.com/matthewdeanmartin/sparkle_log"
__homepage__ = "https://github.com/matthewdeanmartin/sparkle_log"
__documentation__ = "https://matthewdeanmartin.github.io/sparkle_log/sparkle_log/index.html"



"""
Log writer module, stateful and does not require a decorator or context manager.
"""

import logging
import statistics
from typing import Literal, Optional

import psutil

from sparkle_log.drive_space import get_free_percent_for_all_drives
from sparkle_log.graphs import GLOBAL_LOGGER
from sparkle_log.ui import GraphStyle, sparkline

READINGS: dict[str, list[Optional[int]]] = {"cpu": [None] * 29, "memory": [None] * 29, "drive": [None] * 29}

Metrics = Literal["cpu", "memory", "drive"]


def log_system_metrics(metrics: tuple[Metrics, ...], style: GraphStyle = "bar") -> None:
    """
    Log system metrics.

    Args:
        metrics: A tuple of metrics to log.
        style: The style of the sparkline.
    """
    if not GLOBAL_LOGGER.isEnabledFor(logging.INFO):
        return
    if "cpu" in metrics:
        # Interval non to prevent blocking.
        # https://psutil.readthedocs.io/en/latest/#psutil.cpu_percent
        interval = None
        reading = psutil.cpu_percent(interval=interval)

        if reading == 0 and interval is None:
            # First reading on interval None is trash, bail.
            return
        READINGS["cpu"].append(int(reading))
    if "memory" in metrics:
        READINGS["memory"].append(int(psutil.virtual_memory().percent))

    if "drive" in metrics:
        READINGS["drive"].append(int(get_free_percent_for_all_drives()))

    for _key, value in READINGS.items():
        if len(value) > 30:  # Keep the last 30 readings
            value.pop(0)

    for metric, value in READINGS.items():
        if metric not in metrics:
            continue
        values_for_stats = [_ for _ in value if _ is not None]
        average = int(round(statistics.mean(values_for_stats), 0))

        def pad(value: int) -> str:
            """Pad the value with spaces."""
            return str(value).rjust(2)

        minimum = pad(min(values_for_stats))
        maximum = pad(max(values_for_stats))
        metric_formatted = pad(value[-1])[-2:]
        if metric == "cpu":
            GLOBAL_LOGGER.info(
                f"CPU   : {metric_formatted}% "
                f"| min, mean, max ({minimum}, {average}, {maximum}) "
                f"| {sparkline(value, style)}"
            )
        elif metric == "memory":
            GLOBAL_LOGGER.info(
                f"Memory: {metric_formatted}% "
                f"| min, mean, max ({minimum}, {average}, {maximum}) "
                f"| {sparkline(value, style)}"
            )
        elif metric == "drive":
            GLOBAL_LOGGER.info(
                f"Drive: {metric_formatted}% "
                f"| min, mean, max ({minimum}, {average}, {maximum}) "
                f"| {sparkline(value, style)}"
            )



"""
Context manager of logger so you have more control over when it starts and stops.
"""

import logging
import time
from threading import Event, Thread
from typing import Any, Optional

from sparkle_log.graphs import GLOBAL_LOGGER
from sparkle_log.scheduler import run_scheduler
from sparkle_log.ui import GraphStyle


class MetricsLoggingContext:
    """
    Context manager to log system metrics.
    """

    def __init__(self, metrics=("cpu", "memory"), interval: int = 10, style: GraphStyle = "faces") -> None:
        """Initialize the context manager."""
        if not metrics:
            metrics = ("cpu", "memory")
        else:
            for metric in metrics:
                if metric not in ("cpu", "memory", "drive"):
                    raise TypeError("Unexpected metric")
        self.metrics = metrics
        self.interval = interval
        self.style = style
        self.stop_event: Optional[Event] = None
        self.scheduler_thread: Optional[Thread] = None

    def __enter__(self) -> "MetricsLoggingContext":
        """Start the context manager, if logging enabled."""
        if GLOBAL_LOGGER.isEnabledFor(logging.INFO):
            self.stop_event = Event()
            self.scheduler_thread = Thread(
                target=run_scheduler, args=(self.stop_event, self.metrics, self.interval, self.style)
            )
            self.scheduler_thread.start()
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Stop the context manager."""
        if self.scheduler_thread and self.scheduler_thread.is_alive():
            self.stop_event.set()
            self.scheduler_thread.join()


# Usage example with the context manager
if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    with MetricsLoggingContext(metrics=("cpu",), interval=1):
        # Execute some operations
        print("Monitoring system metrics during operations...")
        time.sleep(20)
        # Operations are now being monitored
        # The thread will be stopped when exiting this block



"""Trigger logging on a schedule."""

import time
from threading import Event

import schedule

from sparkle_log.log_writer import log_system_metrics
from sparkle_log.ui import GraphStyle


def run_scheduler(stop_event: Event, metrics: tuple[str, ...], seconds: int, style: GraphStyle = "bar"):
    """Run scheduled tasks until the stop_event is set."""
    schedule.every(seconds).seconds.do(lambda: log_system_metrics(metrics, style))

    while not stop_event.is_set():
        schedule.run_pending()
        time.sleep(1)



"""
Sparkline graph code goes here.
"""

from typing import Literal

import sparklines

GraphStyle = Literal["bar", "faces", "jagged", "linear", "vertical", "ascii_art", "pie_chart"]


def sparkline(numbers: list[int], style: GraphStyle = "bar") -> str:
    """Generate a simple sparkline string for a list of integers."""
    if style == "bar":
        for line in sparklines.sparklines(numbers):
            return line
    elif style == "jagged":
        return jagged_ascii_sparkline(numbers)
    elif style == "vertical":
        return vertical_ascii_sparkline(numbers)
    elif style == "linear":
        return linear_ascii_sparkline(numbers)
    elif style == "ascii_art":
        return ascii_sparkline(numbers)
    elif style == "pie_chart":
        return pie_chart_sparkline(numbers)
    elif style == "faces":
        return faces_sparkline(numbers)
    return ""


def faces_sparkline(data):
    """Generate a sparkline with face emojis."""
    symbols = ["😞", "😐", "😊", "😁"]
    return sparkline_it(data, symbols)


def sparkline_it(data: list[int], symbols: list[str]):
    """Generate a sparkline with the given symbols."""
    noneless_data = [_ for _ in data if _ is not None]
    max_val = max(noneless_data)
    min_val = min(noneless_data)
    range_val = max_val - min_val
    if range_val == 0:  # Avoid division by zero
        return "".join(symbols[0] for _ in data)
    if range_val * len(symbols) == 0:
        return " "
    return "".join(
        symbols[min(len(symbols) - 1, int((val - min_val) / range_val * len(symbols)))] if val is not None else " "
        for val in data
    )


def pie_chart_sparkline(data):
    """Using different geometric shapes or other symbols"""
    symbols = ["○", "◔", "◑", "◕", "●"]
    return sparkline_it(data, symbols)


def ascii_sparkline(data):
    """Using different ASCII characters."""
    symbols = [" ", ".", ":", "-", "=", "+", "*", "#", "%", "@"]
    return sparkline_it(data, symbols)


def linear_ascii_sparkline(data):
    """Ascii for Low, medium, and high levels"""
    levels = ["_", "-", "¯"]
    return sparkline_it(data, levels)


def jagged_ascii_sparkline(data):
    """Ascii incorporating a peak character"""
    symbols = ["_", "-", "^", "¯"]
    return sparkline_it(data, symbols)


def vertical_ascii_sparkline(data):
    """Ascii Using single and double vertical lines"""
    symbols = ["_", "|", "‖"]
    return sparkline_it(data, symbols)



"""
Write sparkline graphs of CPU and memory usage to your logs.
"""

__all__ = ["monitor_metrics_on_call", "log_system_metrics", "MetricsLoggingContext", "__version__"]

from sparkle_log.__about__ import __version__
from sparkle_log.as_context_manager import MetricsLoggingContext
from sparkle_log.as_decorator import monitor_metrics_on_call
from sparkle_log.log_writer import log_system_metrics
